/*****************************************************************************/
/* Simple fire effect demo.                                                  */
/*                                                                           */
/* Original algorithm by Midnight.                                           */
/* Palette tweaks by Sphinx (long live C--!)                                 */
/* Various algorithm tweaks and optimisation, plus port to C and PSP,        */
/* by Fanjita.                                                               */
/*                                                                           */
/* Windows screensaver also available ;)                                     */
/*****************************************************************************/
#include <pspuser.h>
#include <pspctrl.h>
#include <pspdisplay.h>

#include <stdlib.h>
#include <string.h>


#define SCREEN_WIDTH	 512
#define SCREEN_HEIGHT  272
#define DATA_WIDTH     240
#define DATA_HEIGHT		 136
#define LINESIZE       512				//in short
#define FRAMESIZE      0xAA000			//in byte

unsigned char *vramtop=(unsigned char *)0x04000000;
unsigned long drawframe;

PSP_MODULE_INFO("Flamez", 0, 1, 1);

// get vram address for character position
unsigned char *GetVramAddr(unsigned long x,unsigned long y)
{
	return vramtop+(drawframe?FRAMESIZE:0)+x*4+y*LINESIZE*4+0x40000000;
}

void Fillvram(unsigned long color)
{
	unsigned char *vptr0;		
	unsigned long i;

	vptr0=GetVramAddr(0,0);

	for (i=0; i<(FRAMESIZE/4); i++)
	{
		*(unsigned long *)vptr0=color;
		vptr0+=4;
	}
}

void changeBuffer()
{
  sceDisplayWaitVblankStart();

	sceDisplaySetFrameBuf((void*)GetVramAddr(0,0),
                        512,
                      	PSP_DISPLAY_PIXEL_FORMAT_8888,
                        0);

	drawframe=(drawframe?0:1);
}

void initScreenAndCtrl()
{
  drawframe = 0;
	sceDisplaySetMode(0,SCREEN_WIDTH,SCREEN_HEIGHT);

	sceDisplaySetFrameBuf(
      (void*)vramtop,
      512,
    	PSP_DISPLAY_PIXEL_FORMAT_8888
      ,1);
	Fillvram(0x00000000);
  drawframe = 1;
	Fillvram(0x00000000);
  sceCtrlSetSamplingCycle(0);
  sceCtrlSetSamplingMode(1);
}

unsigned int  palette[256];
unsigned char firedata[DATA_HEIGHT + 1][DATA_WIDTH];

#define MAX_ADD_LENGTH   20
void AddFire (unsigned int xicount)
{
	int length;
	int position;
	unsigned char value;

	while (xicount-- > 0)
	{
		length = (rand() % MAX_ADD_LENGTH) + 1;
		position = (rand() % (DATA_WIDTH - MAX_ADD_LENGTH)) ;
		value = (rand() & 0x7F) + 0x80;

		while (length-- > 0)
		{
			firedata[DATA_HEIGHT - 1][position++] = value;
			firedata[DATA_HEIGHT][position] = value;
		}
	}
}

/*****************************************************************************/
/* This simulates the cooling effect as the flames rise, by using a          */
/* weighted average plus non-linear cooling as the heat goes up.             */
/*****************************************************************************/
void CopyAvg ()
{
	unsigned int lrow, lcol;
	unsigned int  llocalsum;

	for (lrow = 2; lrow < DATA_HEIGHT + 1; lrow++)
	{
		for (lcol = 1; lcol < DATA_WIDTH - 1; lcol++)
		{
			llocalsum = firedata[lrow][lcol] +
				          firedata[lrow][lcol-1] +
				          firedata[lrow][lcol+1] +
				          firedata[lrow+1][lcol];
			llocalsum >>= 2;  // average the 4 pixels
			llocalsum += firedata[lrow-1][lcol];
			llocalsum >>= 1;  // average again.

      /***********************************************************************/
      /* Non-linear cooling - more rapid cool-off for lower temps.           */
      /***********************************************************************/
			if (llocalsum > 128)
			{
				llocalsum -= 2;
			}
			else if (llocalsum > 4)
			{
				llocalsum -= 3;
			}
			else
			{
				llocalsum = 0;
			}

			firedata[lrow-2][lcol] = (unsigned char)llocalsum;
		}
	}
}

void initData()
{
	int ii;

  /***************************************************************************/
  /* Set palette up with firey colours, from white for most intense heat     */
  /* through yellow and to red for least.                                    */
  /***************************************************************************/
  palette[0]=0;

	for (ii=0; ii < 85; ii++)
	{
		unsigned char value = (ii * 256) / 85;

		// reds
		palette[ii + 1] = (unsigned int)value;

		// red - yellows
		palette[ii + 86] = (((unsigned int)value) << 8) + 0xFF;

		// yellows - whites
		palette[ii + 171] = (((unsigned int)value) << 16) + 0xFFFFL;
	}

	memset((void*)firedata, 0, (DATA_HEIGHT + 1) * DATA_WIDTH);

  /***************************************************************************/
	/* Stick some initial data into the fire array                             */
	/***************************************************************************/
  ii = 0;
  do
	{
      AddFire((ii >> 5) + 1);
      CopyAvg();
      ii++;
  }
	while( ii < DATA_HEIGHT);
}

SceCtrlData gpaddata;

void transferToScreen()
{
	unsigned int  lrow;
	unsigned int  lcol;
	unsigned int *lpixel;
	unsigned int  lpixelvalue;

	for (lrow = 0; lrow < DATA_HEIGHT - 1; lrow++)
	{
		lpixel = (unsigned int*)GetVramAddr(0, lrow * (SCREEN_HEIGHT / DATA_HEIGHT));

		for (lcol = 0; lcol < DATA_WIDTH; lcol++)
		{
			lpixelvalue = palette[firedata[lrow][lcol]];

			lpixel[SCREEN_WIDTH] = lpixelvalue;
			*lpixel++   = lpixelvalue;
			lpixel[SCREEN_WIDTH] = lpixelvalue;
			*lpixel++   = lpixelvalue;
		}
	}
}

int main()
{
  initScreenAndCtrl();

	initData();

	gpaddata.Buttons = 0xFFFFFFFFL;
	while (gpaddata.Buttons == 0xFFFFFFFFL)
	{
    AddFire(12);
    CopyAvg();

		transferToScreen();
		changeBuffer();
		sceCtrlPeekBufferNegative(&gpaddata, 1);
	}

	return(0);
}
